import copy
from bunch import Bunch
import datetime

from blockerbugs.models.milestone import Milestone
from blockerbugs.models.release import Release
from blockerbugs.models.bug import Bug
from blockerbugs import db
from blockerbugs.util.bug_sync import BugSync

templatebug = Bunch({
    'bug_id': 123456,
    'weburl': 'https://bugzilla.redhat.com/show_bug.cgi?id=123456',
    'component': 'randomcomponent',
    'status': 'NEW',
    'summary': 'this is a test bug',
    'whiteboard': '',
    'last_change_time': datetime.datetime.utcnow(),
    'flags': [{'name': 'needinfo', 'status': '?', 'is_active': True, 'requestee': 'John Doe'}]
})


templatebuginfo = {
    'bug_id': 123456,
    'url': 'https://bugzilla.redhat.com/show_bug.cgi?id=123456',
    'component': 'randomcomponent',
    'status': 'NEW',
    'summary': 'this is a test bug',
    'whiteboard': '',
    'accepted': False,
    'rejected': False,
    'proposed': False,
    'last_change_time': datetime.datetime.utcnow(),
    'active': True,
    'needinfo': 'John Doe'
}


class StubBzInterface(object):
    def __init__(self, deps=[], bugs=[]):
        self.deps = deps
        self.bugs = bugs

    def get_deps(self, tracker_id):
        return self.deps

    def do_query(self, tracker, last_update):
        return self.bugs


class TestfuncBugsync(object):
    @classmethod
    def setup_class(cls):
        db.session.rollback()
        db.drop_all()
        db.create_all()
        test_release17 = Release(17)
        test_milestone17final = Milestone(test_release17, 'final', 752650,
                                          752653, '17-final')
        test_milestone17final.active = False
        db.session.add(test_milestone17final)
        test_milestone17beta = Milestone(test_release17, 'beta', 752649,
                                         752652, '17-beta')
        test_milestone17beta.active = True
        db.session.add(test_milestone17beta)
        db.session.commit()


    @classmethod
    def teardown_class(cls):
        db.session.rollback()
        db.drop_all()

    def setup_method(self, method):
        self.basicbug = copy.deepcopy(templatebug)
        self.basicbuginfo = copy.deepcopy(templatebuginfo)
        self.milestone = Milestone.query.filter_by(active=True).first()

    def teardown_method(self, method):
        oldbugs = Bug.query.all()
        if oldbugs:
            for oldbug in oldbugs:
                db.session.delete(oldbug)
            db.session.commit()
        else:
            print "no old bug found to delete!"


    # test that a bug which is no longer blocking a tracker but not marked as
    # RejectedBlocker is moved to rejected
    def should_removed_tracker_rejected(self):
        tracker_type = 'Blocker'
        self.basicbug.proposed = True
        self.basicbuginfo['proposed'] = True
        testbug = Bug.from_data(self.basicbuginfo, self.milestone,
                                tracker_type)
        db.session.add(testbug)
        db.session.commit()

        deps = []
        stub_interface = StubBzInterface(deps=deps)

        test_sync = BugSync(db, stub_interface)

        test_sync.cleanup_milestone(self.milestone, tracker_type)

        test_syncbugs = Bug.query.filter_by(bugid=self.basicbug.bug_id).all()
        assert len(test_syncbugs) == 1
        test_syncbug = test_syncbugs[0]

        assert test_syncbug.rejected_blocker
        assert not test_syncbug.proposed_blocker
        assert not test_syncbug.accepted_blocker

    # when cleaning up old bugs, bugs not associated with the tracker in question
    # should not be modified
    def should_cleanup_ignore_unassociated(self):
        tracker_type = 'Blocker'
        self.basicbug.proposed = True
        self.basicbuginfo['proposed'] = True
        testbug = Bug.from_data(self.basicbuginfo, self.milestone,
                                tracker_type)
        db.session.add(testbug)

        innocentbuginfo = copy.deepcopy(self.basicbuginfo)
        innocentbuginfo['bug_id'] = 234567
        innocentbuginfo['proposed'] = True
        innocentbug = Bug.from_data(innocentbuginfo, self.milestone,
                                    'FreezeException')
        db.session.add(innocentbug)

        db.session.commit()

        deps = []
        stub_interface = StubBzInterface(deps=deps)

        test_sync = BugSync(db, stub_interface)

        test_sync.cleanup_milestone(self.milestone, tracker_type)

        test_syncbug = Bug.query.filter_by(bugid=innocentbug.bugid).first()

        assert not test_syncbug.rejected_blocker
        assert not test_syncbug.rejected_fe
        assert test_syncbug.proposed_fe

    # when a bug is still blocking the tracker, it should be left unmodified
    def should_cleanup_ignore_blocking(self):
        tracker_type = 'Blocker'
        self.basicbug.proposed = True
        self.basicbuginfo['proposed'] = True
        testbug = Bug.from_data(self.basicbuginfo, self.milestone,
                                tracker_type)
        db.session.add(testbug)
        db.session.commit()

        deps = [self.basicbuginfo['bug_id']]
        stub_interface = StubBzInterface(deps=deps)

        test_sync = BugSync(db, stub_interface)

        test_sync.cleanup_milestone(self.milestone, tracker_type)

        test_syncbug = Bug.query.filter_by(bugid=self.basicbug.bug_id).first()

        assert test_syncbug.proposed_blocker
        assert not test_syncbug.rejected_blocker
        assert not test_syncbug.accepted_blocker

    # when a rejected bug moves is re-added to the tracker w/o RejectedBlocker
    # in the whiteboard, it should be proposed
    def should_readded_blocker_proposed(self):
        tracker_type = 'Blocker'
        self.basicbug.rejected = True
        self.basicbuginfo['rejected'] = True
        testbug = Bug.from_data(self.basicbuginfo, self.milestone,
                                tracker_type)
        db.session.add(testbug)
        db.session.commit()

        self.basicbug.rejected = False
        self.basicbug.proposed = True
        bugs = [self.basicbug]
        stub_interface = StubBzInterface(bugs=bugs)

        test_sync = BugSync(db, stub_interface)

        test_sync.update_milestone(self.milestone, tracker_type)

        test_syncbug = Bug.query.filter_by(bugid=self.basicbug.bug_id).first()

        assert not test_syncbug.rejected_blocker
        assert test_syncbug.proposed_blocker
        assert not test_syncbug.accepted_blocker

    # when a rejected bug from one milestone is re-proposed as a bug for another
    # milestone, a new bug entry should be created for the new milestone
    def should_readded_newmilestone_newbugcreated(self):
        tracker_type = 'Blocker'
        self.basicbug.rejected = True
        self.basicbuginfo['rejected'] = True
        testbug = Bug.from_data(self.basicbuginfo, self.milestone,
                                tracker_type)
        db.session.add(testbug)
        db.session.commit()

        # now add the same bugid to a newer milestone
        newer_milestone = Milestone.query.filter_by(active=False).first()

        self.basicbug.rejected = False
        self.basicbug.proposed = True
        bugs = [self.basicbug]
        stub_interface = StubBzInterface(bugs=bugs)

        test_sync = BugSync(db, stub_interface)

        test_sync.update_milestone(newer_milestone, tracker_type)

        test_syncbugs = Bug.query.filter_by(bugid=self.basicbug.bug_id).all()

        assert len(test_syncbugs) == 2

        # this isn't done yet, need to monkey with the db schema and want to 
        # commit current work first
        test_syncbug = Bug.query.filter_by(bugid=self.basicbug.bug_id,
                                           milestone=newer_milestone).first()

        assert not test_syncbug.rejected_blocker
        assert test_syncbug.proposed_blocker
        assert not test_syncbug.accepted_blocker

    # when a accepted blocker is closed, it should be left alone except for 
    # marking the active feild false
    def should_closed_accepted_inactive(self):
        tracker_type = 'Blocker'
        self.basicbug.accepted = True
        self.basicbuginfo['accepted'] = True
        testbug = Bug.from_data(self.basicbuginfo, self.milestone,
                                tracker_type)
        db.session.add(testbug)
        db.session.commit()

        self.basicbug.status = 'CLOSED'
        bugs = [self.basicbug]
        stub_interface = StubBzInterface(bugs=bugs)

        test_sync = BugSync(db, stub_interface)

        test_sync.update_milestone(self.milestone, tracker_type)

        test_syncbug = Bug.query.filter_by(bugid=self.basicbug.bug_id).first()

        assert not test_syncbug.active

    # when a bug moves from proposed to accepted, it should not be on the
    # proposed list anymore
    def should_proposed_toaccepted_notproposed(self):
        tracker_type = 'Blocker'
        testbug = Bug.from_data(self.basicbuginfo, self.milestone,
                                tracker_type)
        db.session.add(testbug)
        db.session.commit()

        self.basicbug.whiteboard = 'AcceptedBlocker'
        bugs = [self.basicbug]
        stub_interface = StubBzInterface(bugs=bugs)

        test_sync = BugSync(db, stub_interface)

        test_sync.update_milestone(self.milestone, tracker_type)

        test_syncbugs = Bug.query.filter_by(bugid=self.basicbug.bug_id).all()

        assert len(test_syncbugs) == 1

        test_syncbug = test_syncbugs[0]
        assert test_syncbug.accepted_blocker
        assert not test_syncbug.proposed_blocker

    # this will take a LONG time ...
#    def test_bugsync(self):
#        test_bugsync = BugSync(db)
#
#        test_bugsync.update_active()
#
#        bugs = test_release17final.bugs
#        for bug in bugs:
#            print bug
#        assert 1 == 2
